@@ -1,9 +1,10 @@
 import tempfile
 
+HOME_PAGE = "https://www.datacamp.com/"
 LOGIN_URL = "https://www.datacamp.com/users/sign_in"
 LOGIN_DETAILS_URL = "https://www.datacamp.com/api/users/signed_in"
 
-SESSION_FILE = tempfile.gettempdir() + "/.datacamp"
+SESSION_FILE = tempfile.gettempdir() + "/.datacamp.v2"
 
 PROFILE_URL = "https://www.datacamp.com/profile/{slug}"
 COURSE_DETAILS_API = "https://campus-api.datacamp.com/api/courses/{id}/"
